# This stub file should contain the test suite of pyxu.operator.interop.source.from_source().
#
# from_source() is a low-level function used throughout the Pyxu API; as such it is indirectly
# tested through concrete instantiations in:
#    pyxu.operator.linop.base.py
#    pyxu.abc.arithmetic.py
#    (and many others)
#
# Should specific tests be added in the future, they will be inserted here.
